/* 
   hotswap - Supports hotswapping peripherals in portable computers.

   Copyright (C) 2001 Tim Stadelmann

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <mntent.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <linux/hdreg.h>
#include "system.h"

#define max(a,b) (((a) > (b)) ? (a) : (b)) 

static void usage (int status);
static int decode_switches (int argc, char **argv);
static int detect_ide ();
static int rescan_ide ();
static int unregister_ide ();
static int check_mounted_ide ();
static int unmount_ide ();

/* The model string from the IDE device. */
char ide_device_model[40];
/* Whether an IDE device is configured. */
int ide_device_present;

/* The name the program was run with, stripped of any leading path. */
char *program_name;

/* Option flags and variables */

int verbose; /* --verbose */

static struct option const long_options[] =
{
  {"verbose", no_argument, 0, 'v'},
  {"help", no_argument, 0, 'h'},
  {"version", no_argument, 0, 'V'},
  {NULL, 0, NULL, 0}
};


int
main (int argc, char **argv)
{
  int i;
  int error;
  char response[8];

  program_name = argv[0];

  i = decode_switches (argc, argv);

  if (i == argc)
    {
      /* No ACTION argument. */

      /* Interactive mode. */
      
      printf ("\
hotswap %s for notebook module bays using the secondary IDE bus\n\
Copyright 2001 Tim Stadelmann\n\
This program is free software, licensed under the conditions of the\n\
GNU General Public License version 2, or (at your option), any later\n\
version.\n\n\
", VERSION);
      
      /* Check for an IDE device. */
      error = detect_ide ();
      if (error)
	{
	  exit (EXIT_FAILURE);
	}
      
      if (ide_device_present)
	{  
	  printf ("The following IDE device is currently configured:\n%s\n\
Do you want to remove this device? ", ide_device_model);
	  
	  fgets(response, 8, stdin);
	  if (strncmp("yes", response, max (1, strlen (response) - 1)) != 0)
	    {
	      puts ("Aborting.");
	      exit (EXIT_FAILURE);
	    }
	  
	  /* Is the IDE device mounted? */
	  error = check_mounted_ide ();
	  if (error == -1)
	    exit (EXIT_FAILURE);
	  
	  if (error == 1) 
	    {	
	      error = unmount_ide ();
	      
	      if (error)
		exit (EXIT_FAILURE);
	    }
	  
	  /* Unregister the IDE device. */
	  error = unregister_ide ();
	  if (error)
	    exit (EXIT_FAILURE);
	  
	  puts ("You can now remove the device from the module bay.\n");
	}
      else
	{
	  puts ("\
There is currently no IDE device configured.  (Floppy disk drives,\n\
batteries, and 'travel modules' are not managed by this utility.  If\n\
you want to replace such a module, you should remove it now.)\n\
");
	}
      
      printf ("Would you like to insert an IDE device into the module bay? ");
      
      fgets(response, 8, stdin);
      if (strncmp("yes", response, max (1, strlen (response) - 1)) != 0)
	{
	  puts ("Aborting.");
	  exit (EXIT_FAILURE);
	}
      
      printf ("\
Please insert the new device into the module bay and press RETURN.");
      getc (stdin);

      
      /* Rescan the IDE bus. */
    rescan:
      error = rescan_ide ();
      if (error)
	exit (EXIT_FAILURE);
      
      /* Check for an IDE device. */
      error = detect_ide ();
      if (error)
	{
	  exit (EXIT_FAILURE);
	}
      
      if (ide_device_present)
	{  
	  printf ("The following IDE device has been configured successfully:\
\n%s\n", ide_device_model);
	}
      else
	{
	  puts ("\
No IDE device has been found in the module bay.  If you are sure that\n\
the module you want to configure connects to the IDE bus, make sure\n\
that the module is inserted completely.\n\
");
	  
	  printf ("Do you want to try again? ");

	  fgets(response, 8, stdin);
	  if (strncmp("yes", response, strlen (response) - 1) != 0)
	    {
	      puts ("Aborting.");
	      exit (EXIT_FAILURE);
	    }
	  else
	    {
	      goto rescan;
	    }
	}
      
      /* Return successfully. */
      exit (EXIT_SUCCESS);
   }
  
  else
    {
      /* There is an ACTION argument. */

      if (strcmp ("detect-ide", argv[i]) == 0)
	{
	  /* Detect the IDE device. */
	  error = detect_ide ();
	  if (error)
	    {
	      exit (EXIT_FAILURE);
	    }
	  else
	    {
	      printf ("%s\n", ide_device_model);
	      exit (EXIT_SUCCESS);
	    }
	}
      
      else if (strcmp ("mounted-ide", argv[i]) == 0)
	{
	  /* Is the IDE device mounted? */
	  error = check_mounted_ide ();
	  if (error == -1)
	    exit (EXIT_FAILURE);
	  
	  if (error == 1)
	    puts ("yes");
	  else
	    puts ("no");
	  
	  exit (EXIT_SUCCESS);
	}
      
      else if (strcmp ("umount-ide", argv[i]) == 0)
	{
	  /* Is the IDE device mounted? */
	  error = check_mounted_ide ();
	  if (error == -1)
	    exit (EXIT_FAILURE);
	  
	  if (error == 0)
	    {
	      printf ("%s: ", program_name);
	      puts ("The IDE device is not currently mounted.");
	      exit (EXIT_FAILURE);
	    }
	  
	  error = unmount_ide ();
	  
	  if (error)
	    exit (EXIT_FAILURE);
	  else
	    exit (EXIT_SUCCESS);
	}
      
      else if (strcmp ("remove-ide", argv[i]) == 0)
	{
	  /* Is the IDE device mounted? */
	  error = check_mounted_ide ();
	  if (error == -1)
	    exit (EXIT_FAILURE);
	  
	  if (error == 1) 
	    {	
	      printf ("%s: ", program_name);
	      puts ("\
The IDE device is currently mounted and cannot be removed.");
	      exit (EXIT_FAILURE);
	    }
	  
	  /* Unregister the IDE device. */
	  error = unregister_ide ();
	  if (error)
	    exit (EXIT_FAILURE);
	  else
	    exit (EXIT_SUCCESS);
	}
      
      else if (strcmp ("attach-ide", argv[i]) == 0)
	{ 
	  /* Is an IDE device already configured? */
	  error = detect_ide ();
	  if (error)
	    exit (EXIT_FAILURE);
	  if (ide_device_present) 
	    {
	      printf ("%s: ", program_name);
	      printf ("An IDE device is already configured:\n%s\n",
		      ide_device_model);
	      exit (EXIT_FAILURE);
	    }
	  
	  /* Rescan the IDE bus. */
	  error = rescan_ide ();
	  if (error)
	    exit (EXIT_FAILURE);
	  else
	    exit (EXIT_SUCCESS);
	}

      else
	{
	  printf ("%s: ", program_name);
	  printf ("%s: unknown action\n\n", argv[i]);
	  usage (EXIT_FAILURE);
	}
    }
  
}

/* Set all the option flags according to the switches specified.
   Return the index of the first non-option argument. */

static int
decode_switches (int argc, char **argv)
{
  int c;

  while ((c = getopt_long (argc, argv, 
			   "v"	/* verbose */
			   "h"	/* help */
			   "V",	/* version */
			   long_options, (int *) 0)) != EOF)
    {
      switch (c)
	{
	case 'V':
	  printf ("hotswap %s\n", VERSION);
	  exit (EXIT_SUCCESS);
	case 'h':
	  usage (EXIT_SUCCESS);
	case 'v':
	  verbose = 1;
	  break;
	default:
	  usage (EXIT_FAILURE);
	}
    }

  return optind;
}


static void
usage (int status)
{
  printf (_("%s - \
Supports hotswapping peripherals in portable computers.\n"), program_name);
  printf (_("Usage: %s [OPTION]... [ACTION]\n"), program_name);
  printf (_("\
\n\
Options:\n\
  -v, --verbose              print more information\n\
  -h, --help                 display this help and exit\n\
  -V, --version              output version information and exit\n\
\n\
If no ACTION is specified, interactive mode is entered.  In this mode\n\
the program guides through the hotswapping process as appropriate for\n\
the current configuration.\n\
\n\
ACTION can be one of:\n\
  detect-ide                 detect a configured IDE device\n\
  mounted-ide                check whether the IDE device is mounted\n\
  umount-ide                 unmount the IDE device\n\
  remove-ide                 remove the IDE device\n\
  attach-ide                 attach an IDE device\n\
\n\
The program assumes the module bay is connected to the secondary IDE\n\
bus.  The Linux kernel requires hotswappable devices to be the only\n\
device on their IDE bus.\n\
"));
  exit (status);
}


/* Detects whether /dev/hdc is configured and obtains the model
   identification string.  Returns -1 if an error has occured, zero
   otherwise.  The model is written into the global buffer
   ide_device_model; defaults for absent and unidentified devices are
   provided.  The global variable ide_device_present indicates whether
   a configured IDE device is present. */
static int
detect_ide ()
{
  struct hd_driveid identity;
  int device;
  int error;

  if (verbose)
    {
      printf ("%s: ", program_name);
      puts ("Detecting the configured IDE device.");
    }

  ide_device_present = 0;

  /* Open the device.  O_NONBLOCK prevents the open from failing if no
     medium is inserted. */
  device = open ("/dev/hdc", O_RDONLY | O_NONBLOCK);
  if (device < 0)
    {
      /* A failure generally means that no drive is present. */
      strncpy (ide_device_model, "absent", 40);
      return 0;
    }

  /* Get the drive identification. */
  error = ioctl (device, HDIO_GET_IDENTITY, &identity);

  if ((error && (errno == ENOMSG))
      || (!error && (identity.model[0] == '\0')))
    {
      strncpy (ide_device_model, "unidentified IDE device", 40);
      error = 0;
      ide_device_present = 1;
    }
  else if (error)
    {
      printf ("%s: ", program_name);
      perror ("Obtaining IDE drive indentification failed");
    }
  else
    {
      strncpy (ide_device_model, identity.model, 40);
      ide_device_present = 1;
    }

  /* Close the device. */
  close (device);

  return error;
}


/* Rescan the secondary IDE bus.  Returns -1 if an error has occured,
   zero otherwise. */
static int
rescan_ide ()
{
  int device;
  int args[] = {0x170, 0x376, 15};
  int error;

  if (verbose)
    {
      printf ("%s: ", program_name);
      puts ("Rescanning the IDE bus.");
    }

  /* The HDIO_SCAN_HWIF ioctl can be called on any IDE device. */ 
  device = open ("/dev/hda", O_RDONLY);
  if (device < 0) {
    printf ("%s: ", program_name);
    perror ("Opening IDE device failed");
    return -1;
  }

  /* Issue the ioctl. */  
  error = ioctl (device, HDIO_SCAN_HWIF, args);
  if (error != 0)
    {
      if (errno == EIO) /* See ide.c in the kernel. */
	{
	  /* No IDE device found. */
	  error = 0;
	}
      else
	{
	  printf ("%s: ", program_name);
	  perror ("Rescanning the IDE bus failed");
	}
    }
  /* Close the device. */
  close (device);

  return error;
}


/* Unregister the device configured as MASTER on the secondary IDE
   bus.  Returns -1 if an error has occured, zero otherwise. */
static int 
unregister_ide ()
{
  int device;
  int error;

  if (verbose)
    {
      printf ("%s: ", program_name);
      puts ("Unregistering the IDE device.");
    }

  /* The HDIO_UNREGISTER_HWIF ioctl can be called on any IDE
     device. */
  device = open ("/dev/hda", O_RDONLY);
  if (device < 0) {
    printf ("%s: ", program_name);
    perror ("Opening IDE device failed");
    return -1;
  }

  /* Issue the ioctl. */
  error = ioctl (device, HDIO_UNREGISTER_HWIF, 1);
  if (error)
    {
      printf ("%s: ", program_name);
      perror ("Unregistering the IDE device failed");
    }

  /* Close the device. */
  close (device);

  return error;
}


/* Check whether the IDE device or a partition on it is mounted.
   Returns -1 for an error, 1 if at least one mount is active, and 0
   otherwise. */
int
check_mounted_ide ()
{
  FILE *mtab;
  struct mntent *entry;
  int mounted = 0;

  if (verbose)
    {
      printf ("%s: ", program_name);
      puts ("Checking whether the IDE device is mounted.");
    }

  /* Open /etc/mtab. */
  mtab = setmntent ("/etc/mtab", "r");
  
  if (mtab == NULL)
    {
      printf ("%s: ", program_name);
      perror ("Opening /etc/mtab failed");
      return -1;
    }

  /* Loop through the entries in /etc/mtab. */
  entry = getmntent (mtab);
  while (entry != NULL)
    {
      if (strncmp("/dev/hdc", entry->mnt_fsname, 8) == 0)
	{
	  mounted = 1;
	  /* At least one partition is mounted. */
	  break;
	}
      entry = getmntent (mtab);
    }
  
  /* Close /etc/mtab. */
  endmntent (mtab);

  return mounted;
}


int
unmount_ide ()
{
  FILE *mtab, *mtab_tmp;
  struct mntent *entry;
  char buffer[512];
  int error;

  if (verbose)
    {
      printf ("%s: ", program_name);
      puts ("Unmounting the IDE device.");
    }

  /* Open /etc/mtab. */
  mtab = setmntent ("/etc/mtab", "r");
  
  if (mtab == NULL)
    {
      printf ("%s: ", program_name);
      perror ("Opening /etc/mtab failed");
      return -1;
    }

  /* Loop through the entries in /etc/mtab. */
  entry = getmntent (mtab);
  while (entry != NULL)
    {
      if (strncmp("/dev/hdc", entry->mnt_fsname, 8) == 0)
	{
	  if (verbose)
	    {
	      printf ("%s: ", program_name);	      
	      printf ("Unmounting %s.\n", entry->mnt_dir);
	    }
	  error = umount (entry->mnt_dir);
	  if (error < 0)
	    {
	      printf ("%s: ", program_name);
	      perror ("Unmounting failed");
	      return error;
	    }
	}
      entry = getmntent (mtab);
    }

  /* Remove entries from /etc/mtab. */

  /* Go back to the beginning of /etc/mtab. */
  rewind (mtab);

  /* Open a temporary file for writing. */
  mtab_tmp = fopen ("/etc/mtab.tmp", "w");
  
  if (mtab_tmp == NULL)
    {
      printf ("%s: ", program_name);
      perror ("Opening /etc/mtab.tmp failed");
      /* Close /etc/mtab. */
      endmntent (mtab);
      return -1;
    }

  while (fgets (buffer, 512, mtab) != NULL)
    {
      if (strncmp("/dev/hdc", buffer, 8) != 0)
	fputs (buffer, mtab_tmp);
    }

  /* Close /etc/mtab. */
  endmntent (mtab);  

  /* Close /etc/mtab.tmp. */
  fclose (mtab_tmp);

  /* Replace /etc/mtab by the temporary file. */
  error = rename ("/etc/mtab.tmp", "/etc/mtab");

  return 0;
}
